#! /usr/bin/perl -w

use strict;

my $sender_regexp = '[^:]+: ([^:]+): from=<([^>]+)>, size=([0-9]+), ';
my $delivery_regexp = '[^:]+: ([^:]+): to=<([^>]+)>, ';

my $result = 0; # SUCCESS
my $error_reason = "unknown";

my %messages;

sub dump_messages($$);

# BUGS
#    - only the last error will be shown

while (my $line = <STDIN>) {
   if ($line =~ /$sender_regexp/) {
      my ($message_id, $from, $size) = ($1, $2, $3);

      if (exists($messages{$message_id})) {
         $result = 1; # FAILURE
         $error_reason = 'duplicate Message-ID or wrong sequencing';
      }
      else {
         $messages{$message_id} = { 'from' => $from,
                                    'size' => $size,
                                    'to' => []
                                  };
      }
   }
   elsif ($line =~ /$delivery_regexp/) {
      my ($message_id, $to) = ($1, $2);

      if (!exists($messages{$message_id})) {
         $result = 1; # FAILURE
         $error_reason = 'no matching Message-ID or wrong sequencing';
      }
      else {
         push(@{$messages{$message_id}->{'to'}},
                          $to);
      }
   }
   else {
      print STDERR "tossing: ", $line, "\n";
      $result = 1; # FAILURE
      $error_reason = 'tossed unparseable lines';
   }
}

if ($result == 0) {
   my $msg = dump_messages(\%messages, \$error_reason);
   if (defined($msg)) {
      print $msg, "\n";
   }
   else {
      $result = 1; # FAILURE
      $error_reason = 'dump_messages() failed: ' . $error_reason;
   }
}

if ($result != 0) {
   print STDERR $0, ": failed: ", $error_reason, "\n";
}

exit $result;

sub dump_entry($$$) {
   my ($entry_ref, $spaces, $error_reason_ref) = @_;

   my $str = '';

   foreach my $key (keys %{$entry_ref}) {
      $str .= $spaces . $key . ': ';

      if ($key eq 'to') {
         $str .= join(', ', @{$entry_ref->{$key}});
      }
      else {
         $str .= $entry_ref->{$key};
      }

      $str .= "\n";
   }

   return $str;
}

sub dump_messages($$) {
   my ($messages_hash_ref, $error_reason_ref) = @_;

   my $str = '';

   while (my ($key, $value) = each(%{$messages_hash_ref})) {
      $str .= 'message-id: ' . $key . "\n";

      my $tmp = dump_entry($value, '   ', $error_reason_ref);
      if (defined($tmp)) {
         $str .= $tmp;
      }
      else {
         $$error_reason_ref = 'failed with '
                              . $key
                              . ': '
                              . $$error_reason_ref;
         return undef;
      }

      $str .= "\n";
   }

   return $str;
}