package generic_storage;
use strict;
use gppds;

# Class variables
my $object_count = 0;

sub new {
   my ($class, %parameters) = @_;

   my $error_reason;

   my @mandatory_parameters = ('hostname',
                               'dbname',
                               'tablename',
                               'username',
                               'password',
                               'identifiant',
                               'value');

   my $ok = 1;
   foreach my $p (@mandatory_parameters) {
      if (!defined($parameters{$p})) {
         $ok = 0;
      }
   }

   if ($ok) {
      my $dbh = open_database($parameters{'dbname'},
                              $parameters{'hostname'},
                              $parameters{'username'},
                              $parameters{'password'},
                              \$error_reason);
      if (defined($dbh)) {
         $object_count++;

	 return bless { 'dbh' => $dbh,
			'tablename' => $parameters{'tablename'},
                        'identifiant' => $parameters{'identifiant'},
                        'value' => $parameters{'value'}
		      },
		      $class;
      }
   }

   return undef;
}

sub DESTROY {
   my $self = shift;

   close_database(\$self->{'dbh'});
   delete $self->{'dbh'};

   $object_count--;
}

# Class Method
sub get_object_count {
   return $object_count;
}

# Instance Methods

sub add {
   my ($self, $id, $fname) = @_;

   my $error_reason;

   return do_sql_insert($self->{'dbh'},
                        0,
                        $self->{'tablename'},
                        { $self->{'identifiant'} => $id,
                          $self->{'value'}       => $fname
                        },
                        undef,
                        \$error_reason);
}

sub retrieve {
   my ($self, $id) = @_;

   my $error_reason;

   my $query = "SELECT " . $self->{'value'} . " FROM " . $self->{'tablename'}
               . " WHERE (" . $self->{'identifiant'} ." = ?)";

   my @titles;
   my @content;
   if (do_query($self->{'dbh'},
                $query,
                \@titles,
                \@content,
                [ $id ],
                undef,
                \$error_reason)) {
      # Assumed unique
      if (scalar(@content) == 1) {
         return $content[0]->{$self->{'value'}};
      }
   }

   return undef;
}

sub remove {
   my ($self, $id) = @_;

   my $error_reason;

   if (do_query($self->{'dbh'},
                "DELETE FROM " . $self->{'tablename'}
                . " WHERE (" . $self->{'identifiant'} . " = ?)",
                undef,
                undef,
                [ $id ],
                undef,
                \$error_reason)) {
      return 1;
   }

   return 0;
}

sub rewind {
   my ($self) = @_;

   if (exists($self->{'in_error'})) {
      delete $self->{'in_error'};
   }

   if (exists($self->{'eof'})) {
      delete $self->{'eof'};
   }

   if (exists($self->{'position'})) {
      delete $self->{'position'};
   }
}

sub in_error {
   my ($self) = @_;

   return defined($self->{'in_error'}) ? 1 : 0;
}

sub eof {
   my ($self) = @_;

   return defined($self->{'eof'}) ? 1 : 0;
}

# BUGS
#    - cursors would be better
sub next {
   my ($self) = @_;

   my $error_reason;

   my $query = <<EOF;
SELECT $self->{'identifiant'}, $self->{'value'} FROM $self->{'tablename'}
%PLACEHOLDER%
ORDER BY $self->{'identifiant'} LIMIT 1
EOF

   my @values;

   if (defined($self->{'position'})) {
      $query =~ s/%PLACEHOLDER%/WHERE ($self->{'identifiant'} > ?)/;
      push(@values, $self->{'position'});
   }
   else {
      $query =~ s/%PLACEHOLDER%//;
   }

   my @titles;
   my @content;
   if (do_query($self->{'dbh'},
                $query,
                \@titles,
                \@content,
                \@values,
                undef,
                \$error_reason)) {
      # Assumed unique
      if (scalar(@content) == 1) {
         $self->{'position'} = $content[0]->{$self->{'identifiant'}};
         return ($content[0]->{$self->{'identifiant'}},
                 $content[0]->{$self->{'value'}});
      }
      else {
         $self->{'eof'} = 1;
         return (undef, undef);
      }
   }

   $self->{'in_error'} = 1;
   return (undef, undef);
}

1;
