#! /usr/bin/perl -w

# Le but de cette application WWW, compose de 3 crans (formulaires)
# diffrents est d'illustrer quelques concepts de la mthode CGI.
#    1er cran:
#       - nom et prnom
#    2me cran
#       - sexe
#    3me cran
#       - date de naissance
#    4me cran
#       - rsum des informations entres.
# BUGS
#    - tests d'intgrit / existence, notamment de session
#    - le choix de mode pourrait tre bas galement sur les donnes
#      remplies, sinon risque de bypass. Et/ou stocker dans la session.

# ------------------------------------------------------------------------

package mon_application;
use strict;
use base 'CGI::Application';
use CGI::Session;

my @screens = ('nom_prenom', 'sexe', 'date_naissance', 'resume');

# Construction d'un hash dont les cls ET les valeurs viennent de
# @screens.
my %run_modes;
@run_modes{@screens} = @screens;

my %next_run_mode;
@next_run_mode{@screens} = (@screens[1..$#screens], undef);

my %values_to_expect_and_store;
@values_to_expect_and_store{@screens}
   = (undef, [ 'nom_prenom' ], [ 'sexe' ], [ 'date_naissance' ]);

sub setup {
   my $self = shift;

   $self->start_mode($screens[0]);
   $self->mode_param('next_mode');
   $self->run_modes(%run_modes);

   my $sid;

   if (defined($self->query()->param('session_id'))) {
      $sid = $self->query()->param('session_id');
   }

   # Initialize session
   my $session = new CGI::Session("driver:File", $sid, {Directory=>'/tmp'});
   $session->expire('+2m');

   $self->param('session' => $session);

   sub teardown {
      my $self = shift;

      # close the session without killing it.
   }

   sub store_cgi_param_in_session {
      my ($self, @params) = @_;

      my $q = $self->query();

      # BUGS
      #    - no constraints
      #    - no way to tell error
      foreach my $p (@params) {
         if (defined($q->param($p))) {
            $self->param('session')->param($p, $q->param($p));
         }
      }
   }

   sub store_session {
      my $self = shift;

      if (defined($values_to_expect_and_store{$self->get_current_runmode()})) {
         $self->store_cgi_param_in_session
            (@{$values_to_expect_and_store{$self->get_current_runmode()}});
      }
   }

   sub next_mode {
      my $self = shift;

      if (defined($next_run_mode{$self->get_current_runmode()})) {
         return $next_run_mode{$self->get_current_runmode()};
      }
      else {
         return undef;
      }
   }

   sub next_mode_form {
      my $self = shift;

      my $q = $self->query();
      my $next_mode = $self->next_mode();
      $q->delete('next_mode');
      if (defined($next_mode)) {
	 return $q->hidden(-name => 'next_mode',
			   -value => $next_mode)
                . $q->hidden(-name => 'session_id',
                             -value => $self->param('session')->id());
      }
      else {
         return '';
      }
   }

   # -------------------------------------------------------------- FORM 1
   sub nom_prenom {
      my $self = shift;

      # Get CGI query object
      my $q = $self->query();

      my $output = '';

      $output .= $q->start_html(-title => 'Entrez vos informations (1/3)');
      $output .= $q->start_form();
      $output .= "Nom et prnom ";
      $output .= $q->textfield(-name => 'nom_prenom');
      $output .= $self->next_mode_form();
      $output .= $q->submit();
      $output .= $q->end_form();
      $output .= $q->end_html();

      return $output;
   }

   # -------------------------------------------------------------- FORM 2
   sub sexe {
      my $self = shift;

      # Get CGI query object
      my $q = $self->query();

      my $output = '';

      $self->store_session();
 
      $output .= $q->start_html(-title => 'Entrez vos informations (2/3)');
      $output .= $q->start_form();
      $output .= "Sexe ";
      $output .= $q->popup_menu(-name => 'sexe',
                                -values => [ 'homme', 'femme' ]),
      $output .= $self->next_mode_form();
      $output .= $q->submit();
      $output .= $q->end_form();
      $output .= $q->end_html();

      return $output;
   }

   # -------------------------------------------------------------- FORM 3

   sub date_naissance {
      my $self = shift;

      # Get CGI query object
      my $q = $self->query();

      my $output = '';

      $self->store_session();

      $output .= $q->start_html(-title => 'Entrez vos informations (3/3)');
      $output .= $q->start_form();
      $output .= "Date de naissance ";
      $output .= $q->textfield(-name => 'date_naissance'),
      $output .= $self->next_mode_form();
      $output .= $q->submit();
      $output .= $q->end_form();
      $output .= $q->end_html();

      return $output;
   }

   # -------------------------------------------------------------- FORM 4

   sub resume {
      my $self = shift;

      # Get CGI query object
      my $q = $self->query();

      my $output = '';

      $self->store_session();

      $output .= $q->start_html(-title => 'Resume');

      my $param_hashref = $self->param('session')->param_hashref();
      foreach my $p (keys %{$param_hashref}) {
         $output .= $q->p 
                    . CGI::escapeHTML($p) . ' -> '
                    . CGI::escapeHTML($param_hashref->{$p});
      }

      $output .= $q->end_html();

      # kill the session
      $self->param('session')->delete();

      return $output;
   }

   1;
}

# ------------------------------------------------------------------------

use strict;
#use mon_application;

my $mon_application = mon_application->new();
$mon_application->run();