#! /usr/bin/perl -w -I/usr/lib/cril-perl-support/lib -I/usr/lib/gppds -I/usr/lib/perl-sessions/lib -I/usr/lib/file-server-admin/lib -I/usr/lib/cgi-support/lib
# file-server-admin Project
# file-server-admin.pl -- Main WWW engine.
# AUTHOR
#    Marc SCHAEFER <schaefer@alphanet.ch>
# LICENSE
#    (C) 2004 by Marc SCHAEFER, licensed under the terms of the
#    GNU General Public License as published by the Free Software
#    Foundation; either version 2 or later (at your option).
# DESCRIPTION
#    This script has, for now, a single function: manage a Samba domain of
#    workstations, groups and users.
# NOTES
#    - Use sessions everywhere, e.g. in order to prevent multiple submits?
#      (ie a session dies after create for example)
#    - A superuser is a superuser if he is a member of the admin-group as
#      set in config. It is superuser for the duration of his session.
# BUGS
#    - Changing groups is inefficient, and will add the user to its
#      primary group in /etc/group, too. See project NOTES.
# TODO
# BASED-ON
#    mail-admin
# MODIFICATION-HISTORY
#    2004-09-23  schaefer  File creation
# $Id: file-server-admin.pl,v 1.15 2005/04/18 15:03:12 schaefer Exp $

use strict;
use CGI; # qw(-nosticky);
use URI::Escape;
use HTML::Entities;
use gppds; # Database access
use perl_sessions; # sessions
use smb_password_handling;
use cgi_forms;
use cgi_menus;
use cgi_results;
use config; # Configuration
use list;
use piped_commands;
use file_server_admin; # Configuration

my $result = 0; # SUCCESS
my $error_reason = 'unknown';

# HACK HACK HACK
# BUGS
#    - This seems to be required so that errors from DBI do not clobber
#      stdout and thus corrupt the file type. We want the errors in stderr
#      but they apparently also happen in stdout.
#      Probably the correct fix would be PrintError still on in DBI but
#      a handler which would only print to stderr.
#      See also the start_html() function.
local(*OLDSTDOUT);
open(OLDSTDOUT, ">&STDOUT"); # err. ign.
open(STDOUT, '>', "/dev/null");
# END HACK HACK HACK

my $debug = 0;

my $home_directory = '../..'; # assumes ran from there.

my $config_file = $home_directory . '/config';

my $my_templates = '/etc/file-server-admin/templates';

my $printed_header = 0;

my $session_validity = 15 * 60; # 15 minutes

my $html_directory = '/etc/file-server-admin/html';
my %html_files = ('header' => $html_directory . '/' . 'header.html',
                  'header-end' => $html_directory . '/' . 'header_end.html',
                  'footer' => $html_directory . '/' . 'footer.html');

my @additional_session_information = ('is_superuser');
my %session_information;
my $session_resume_query
   = 'SELECT is_superuser, identifiant '
     . 'FROM session s '
       . 'WHERE (s.session_id = ?) '
       . 'AND ((CURRENT_TIMESTAMP - s.session_last_revival) <= '
       . $session_validity
       . ')';

my $windows_user_group = group_get_gid('winusers');
my $windows_machine_group = group_get_gid('winmachi');
my $minimal_uid = 1000;

my $minimal_gid = 1000;
my @exclude_groups = ('fsadmin', 'nogroup', 'nossh', 'winmachi', 'winusers');

my $myself = "file-server-admin.pl";

$| = 1;

my $query = new CGI;

my %config_hash;
if (load_config($config_file,
                \%config_hash,
                \$error_reason,
                undef,
                [ @file_server_admin::required_configs,
                  'command-manage-smb-user',
                  'command-manage-smb-group',
                  'command-manage-smb-machine'
                ])) {
   my $dbh = open_database($config_hash{'database-dbname'},
                           $config_hash{'database-host'},
                           $config_hash{'database-user'},
                           $config_hash{'database-password'},
			   \$error_reason);
   if (defined($dbh)) {
      if (defined($session_information{'identifiant'})) {
         delete $session_information{'identifiant'};
      }
      $session_information{'is_superuser'} = 0; # by default normal user

      my $dont_clean = 0; # clean FORM parameters (no sticky)

      my %login_struct
	 = (1 => { 'name' => 'login',
                   'descr' => 'Utilisateur SMB',
                   'type' => 'textfield',
                   'check' => \&form_valid_param
                 },
            2 => { 'name' => 'password',
		   'descr' => 'Mot de passe',
		   'type' => 'password',
		   'check' => \&form_valid_param
		 });

      my %login_menu_struct
	 = (1 => { 'descr' => 'S\'identifier',
		   'type' => 'form',
		   'what' => \%login_struct,
		   'action' => $myself,
		   'params' => { 'mode' => 'login' }
		 });

      my %user_menu_struct # will append session
	 = (3 => { 'descr' => 'Changement mot de passe',
		   'type' => 'link',
		   'action' => $myself,
		   'params' => { 'commande' => 'user_chg_pw' }
		 },
            4 => { 'descr' => 'Dconnexion',
		   'type' => 'link',
		   'action' => $myself,
		   'params' => { 'commande' => 'logout' }
		 });

      my %admin_menu_struct # will append session
	 = (100 => { 'descr' => 'Gestion des utilisateurs',
		     'type' => 'link',
		     'action' => $myself,
		     'params' => { 'commande' => 'liste_utilisateurs' }
 		   },
	    200 => { 'descr' => 'Gestion des groupes',
		     'type' => 'link',
		     'action' => $myself,
		     'params' => { 'commande' => 'liste_groupes' }
		   },
	    300 => { 'descr' => 'Gestion des machines du domaine',
		     'type' => 'link',
		     'action' => $myself,
		     'params' => { 'commande' => 'liste_machines' }
		   });

      my %samba_user_add_struct
	 = (1 => { 'name'  => 'login',
		   'descr' => 'Login SMB',
		   'type'  => 'textfield',
                   'specification' => 'user',
		   'check' => \&form_valid_user_machine_or_group
		 },
            2 => { 'name'  => 'password',
		   'descr' => 'Mot de passe',
		   'type'  => 'password',
		   'check' => \&form_valid_param
		 });

      my %samba_user_modify_struct
	 = (2 => { 'name'  => 'password',
		   'descr' => 'Mot de passe',
		   'type'  => 'password',
		   'check' => \&form_valid_param
		 });

      my %samba_machine_add_struct
	 = (1 => { 'name'  => 'machine_name',
		   'descr' => 'Nom de machine',
		   'type'  => 'textfield',
                   'specification' => 'machine',
		   'check' => \&form_valid_user_machine_or_group
		 });

      my %samba_group_add_struct
	 = (1 => { 'name'  => 'group_name',
		   'descr' => 'Nom de groupe',
		   'type'  => 'textfield',
                   'specification' => 'groupe',
		   'check' => \&form_valid_user_machine_or_group
		 });

      my %password_change_struct
	 = (1 => { 'name'  => 'new_password',
		   'descr' => 'Nouveau mot de passe',
		   'type'  => 'password',
		   'check' => \&form_valid_param,
		 },
	    2 => { 'name'  => 'new_password_same',
		   'descr' => 'Confirmation du nouveau mot de passe',
		   'type'  => 'password',
		   'check' => \&form_valid_param,
		 });   

      # Completely dynamic!
      my %samba_group_modify_struct;

      my %the_forms = ('login_struct' => \%login_struct,
		       'samba_user_add_struct' => \%samba_user_add_struct,
		       'samba_user_modify_struct'
                          => \%samba_user_modify_struct,
		       'samba_group_add_struct' => \%samba_group_add_struct,
		       'samba_group_modify_struct'
                          => \%samba_group_modify_struct,
                       'samba_machine_add_struct'
                          => \%samba_machine_add_struct,
                       'password_change_struct'
                          => \%password_change_struct);

      my %context = ('user_menu_struct_ref' => \%user_menu_struct,
		     'admin_menu_struct_ref' => \%admin_menu_struct,
		     'myself' => $myself,
                     'session_information' => \%session_information);

      if (defined($query->param('session'))) {
	 $context{'session_id'} = scalar($query->param('session'));

	 if (!session_resume($dbh,
		 	     $context{'session_id'},
			     $session_resume_query,
			     \%session_information,           
			     \$error_reason)) {
	    $error_reason = 'session invalide: ' . $error_reason;
	    $result = 1;
	    delete $session_information{'identifiant'}; # to be safe
	 }
      }

      my $message;

      my @create_form;
      my %additional_form_parameters;

      if (defined($session_information{'identifiant'})) {
	 if (defined($query->param('commande'))) {
	    my $choix = $query->param('commande');

	    if ($choix eq 'logout') {
	       if (session_terminate($dbh,
				     $context{'session_id'},
				     \$error_reason)) {
		  $message = 'Session termine';
	       }
	       else {
		  $error_reason = 'session ne peut tre termine: '
				  . $error_reason;
		  $result = 1;
	       }

	       delete $session_information{'identifiant'};
               $session_information{'is_superuser'} = 0;
	       delete $context{'session_id'};
	    }
            elsif ($choix eq 'user_chg_pw') {
               $message = "Changement de votre mot de passe";
               @create_form = ('password_change_struct');
            }
            elsif ($choix eq 'handle_password_change_struct') {
	       my %param_hash;
	       my @incorrect_fields;
	       if (form_validate_input($query,
				       \%password_change_struct,
				       \%param_hash,
				       \@incorrect_fields)) {
                  if ($param_hash{'new_password'}
                      eq $param_hash{'new_password_same'}) {
                     if (user_update_password($session_information
                                                 {'identifiant'},
                                              $param_hash{'new_password'},
                                              \$error_reason)) {
			$message
			  = 'Mot de passe modifi';
                     }
                     else {
			$result = 1;
			$error_reason = 'Impossible de changer le mot de passe'
                                        . ': '
					. $error_reason;
                     }
                  }
                  else {
                     $result = 1;
                     $error_reason = 'Mots de passe diffrents';
                  }
	       }
	       else {
		  $result = 1;
		  $error_reason = 'Champs incorrects: '
				  . join(', ', @incorrect_fields);
		  @create_form = ('password_change_struct');
		  $dont_clean = 1; # don't FORM parameters except hidden
	       }
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'utilisateur_ajouter')) {
               $message = "Ajout d'un utilisateur";
               @create_form = ('samba_user_add_struct');
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'utilisateur_modifier')
                   && defined($query->param('id'))) {
               my $user = $query->param('id');
	       my $users_ref = users_list([],
					  $windows_user_group,
					  $minimal_uid,
					  \$error_reason);
	       if (list_exists($user, @{$users_ref})) {
		  $message = "Modification de l'utilisateur " . $user;

                  $additional_form_parameters{'id'} = $user;

		  @create_form = ('samba_user_modify_struct');
               }
               else {
		  $result = 1;
		  $error_reason = 'Utilisateur inexistant';
               }
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'handle_samba_user_modify_struct')
                   && defined($query->param('id'))) {
               my $user = $query->param('id');
	       my $users_ref = users_list([],
					  $windows_user_group,
					  $minimal_uid,
					  \$error_reason);
	       if (list_exists($user, @{$users_ref})) {
		  my %param_hash;
		  my @incorrect_fields;
		  if (form_validate_input($query,
					  \%samba_user_modify_struct,
					  \%param_hash,
					  \@incorrect_fields)) {
                     if (user_update_password($user,
                                              $param_hash{'password'},
                                              \$error_reason)) {
                        $message = 'Mot de passe modifi';
                     }
                     else {
			$result = 1;
			$error_reason = 'Erreur lors du changement: '
                                        . $error_reason;
                     }
		  }
		  else {
		     $result = 1;
		     $error_reason = 'Champs incorrects: '
				     . join(', ', @incorrect_fields);
		     @create_form = ('samba_user_modify_struct');
		     $dont_clean = 1;
		  }
               }
               else {
		  $result = 1;
		  $error_reason = 'Utilisateur inexistant';
               }
	    }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'handle_samba_user_add_struct')) {
	       my %param_hash;
	       my @incorrect_fields;
	       if (form_validate_input($query,
				       \%samba_user_add_struct,
				       \%param_hash,
				       \@incorrect_fields)) {
                  my $user = $param_hash{'login'};
                  my $password = $param_hash{'password'};

                  if (user_add($user, $password, \$error_reason)) {
                     $message = 'Utilisateur cr';
                  }
                  else {
                     $result = 1;
                     $error_reason = 'Erreur lors de la cration: '
                                     . $error_reason;
                  }
               }
	       else {
		  $result = 1;
		  $error_reason = 'Champs incorrects: '
				  . join(', ', @incorrect_fields);
                  @create_form = ('samba_user_add_struct');
                  $dont_clean = 1;
	       }
	    }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'groupe_ajouter')) {
               $message = "Ajout d'un groupe";
               @create_form = ('samba_group_add_struct');
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'handle_samba_group_add_struct')) {
	       my %param_hash;
	       my @incorrect_fields;
	       if (form_validate_input($query,
				       \%samba_group_add_struct,
				       \%param_hash,
				       \@incorrect_fields)) {
                  my $group = $param_hash{'group_name'};

                  if (group_add($group, \$error_reason)) {
                     $message = 'Groupe cr';
                  }
                  else {
                     $result = 1;
                     $error_reason = 'Erreur lors de la cration: '
                                     . $error_reason;
                  }
               }
	       else {
		  $result = 1;
		  $error_reason = 'Champs incorrects: '
				  . join(', ', @incorrect_fields);
                  @create_form = ('samba_group_add_struct');
                  $dont_clean = 1;
	       }
	    }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'machine_ajouter')) {
               $message = "Ajout d'une machine";
               @create_form = ('samba_machine_add_struct');
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'handle_samba_machine_add_struct')) {
	       my %param_hash;
	       my @incorrect_fields;
	       if (form_validate_input($query,
				       \%samba_machine_add_struct,
				       \%param_hash,
				       \@incorrect_fields)) {
                  my $machine = $param_hash{'machine_name'};

                  if (machine_add($machine, \$error_reason)) {
                     $message = 'Machine cre';
                  }
                  else {
                     $result = 1;
                     $error_reason = 'Erreur lors de la cration: '
                                     . $error_reason;
                  }
               }
	       else {
		  $result = 1;
		  $error_reason = 'Champs incorrects: '
				  . join(', ', @incorrect_fields);
                  @create_form = ('samba_machine_add_struct');
                  $dont_clean = 1;
	       }
	    }
            # confirmation
            elsif ($session_information{'is_superuser'}
                   && ($choix =~ /_supprimer_c$/)
                   && defined($query->param('id'))) {
               my $param = $query->param('id');
               my $choix_confirme = $choix;
               my $choix_non_confirme = $choix;

               $choix_confirme =~ s/_c$//;
               $choix_non_confirme =~ s/^(.+)_supprimer_c$/liste_$1s/;

               my $title = 'Confirmer la suppression de ' . $param;
	       start_html('text/html', $title, \%context);
	       $printed_header = 1;

               print $query->h1($title), $query->p;

               # Ceci n'est pas vraiment trs beau. Amliorer cgi_support?
               print '<FORM ACTION="', $myself, '">',
                     '<INPUT TYPE="hidden" NAME="session" VALUE="',
                     encode_entities($context{'session_id'}),
                     '">',
                     '<INPUT TYPE="hidden" NAME="id" VALUE="',
                     encode_entities($param),
                     '">',
                     '<INPUT TYPE="hidden" NAME="commande" VALUE="',
                     encode_entities($choix_confirme),
                     '">',
                     '<INPUT TYPE="submit" NAME="submit" VALUE="Confirmer">',
                     '</FORM>';

               print '<FORM ACTION="', $myself, '">',
                     '<INPUT TYPE="hidden" NAME="session" VALUE="',
                     encode_entities($context{'session_id'}),
                     '">',
                     '<INPUT TYPE="hidden" NAME="commande" VALUE="',
                     encode_entities($choix_non_confirme),
                     '">',
                     '<INPUT TYPE="submit" NAME="submit" VALUE="Annuler">',
                     '</FORM>';
            }
            elsif ($session_information{'is_superuser'}
                   && ($choix eq 'utilisateur_supprimer')
		   && defined($query->param('id'))) {
               my $login = $query->param('id');

               if (delete_user($login, \$error_reason)) {
                  $message = 'Utilisateur supprim';
               }
               else {
                  $result = 1;
                  $error_reason = 'Erreur de suppression: ' . $error_reason;
               }
            }
            elsif ($session_information{'is_superuser'}
                   && ($choix eq 'groupe_supprimer')
		   && defined($query->param('id'))) {
               my $group = $query->param('id');

               if ($group eq $config_hash{'admin-group'}) {
                  $error_reason
                     = "Le groupe d'administration ne peut tre supprim";
                  $result = 1;
               }
               else {
		  if (delete_group($group, \$error_reason)) {
		     $message = 'Groupe supprim';
		  }
		  else {
		     $result = 1;
		     $error_reason = 'Erreur de suppression: ' . $error_reason;
		  }
               }
            }
            elsif ($session_information{'is_superuser'}
                   && ($choix eq 'machine_supprimer')
		   && defined($query->param('id'))) {
               my $group = $query->param('id');

               if (delete_machine($group, \$error_reason)) {
                  $message = 'Machine supprime';
               }
               else {
                  $result = 1;
                  $error_reason = 'Erreur de suppression: ' . $error_reason;
               }
            }
	    elsif ($session_information{'is_superuser'}
                    && ($choix eq 'liste_utilisateurs')) {
               # all from the special group for smb users
               # can edit/delete
               # then menu to add

               my $title = 'Liste des utilisateurs';
	       start_html('text/html', $title, \%context);
	       $printed_header = 1;

               my $users_ref = users_list([],
					  $windows_user_group,
					  $minimal_uid,
					  \$error_reason);
               if (defined($users_ref)) {
                  my @titles = ('login');
                  my @content;

                  foreach (sort @{$users_ref}) {
                     push (@content, { 'login' => $_ });
                  }

		  my @url_actions
		     = ({ 'descr' => "modifier",
			  'action_url' => $myself
					  . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=utilisateur_modifier',
			  'id' => 'login'
			},
			{ 'descr' => "supprimer",
			  'action_url' => $myself
					  . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=utilisateur_supprimer_c',
			  'id' => 'login'
			});

		  results_print($title,
				\@content,
				\@titles,
				undef,
				$myself,
				undef,
				undef,
				undef,
				\@url_actions,
				$query);

                  print $query->p,
                        $query->a({href=> $myself
                                          . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=utilisateur_ajouter'},
                                  "Ajouter un utilisateur");
               }
               else {
		  $error_reason = 'liste utilisateurs impossible: '
				  . $error_reason;
		  $result = 1;
               }
	    }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'liste_groupes')) {
               # all in group, except list (configurable?)
               # delete, edit (basically mass remove or add from users,
               # so have functions!)
               # then menu to add
               my $title = 'Liste des groupes';
	       start_html('text/html', $title, \%context);
	       $printed_header = 1;

               my $groupes_ref = group_list(\@exclude_groups,
					    $minimal_gid,
					    \$error_reason);
               if (defined($groupes_ref)) {
		  my @titles = ('groupe');
                  my @content;

                  foreach (sort @{$groupes_ref}) {
                     push (@content, { 'groupe' => $_ });
                  }

		  my @url_actions
		     = ({ 'descr' => "modifier",
			  'action_url' => $myself
					  . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=groupe_modifier',
			  'id' => 'groupe'
			},
                        { 'descr' => "supprimer",
			  'action_url' => $myself
					  . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=groupe_supprimer_c',
			  'id' => 'groupe'
			});

		  results_print($title,
				\@content,
				\@titles,
				undef,
				$myself,
				undef,
				undef,
				undef,
				\@url_actions,
				$query);

                  print $query->p,
                        $query->a({href=> $myself
                                          . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=groupe_ajouter'},
                                  "Ajouter un groupe");
               }
               else {
		  $error_reason = 'liste groupes impossible: '
				  . $error_reason;
		  $result = 1;
               }
	    }
            elsif ($session_information{'is_superuser'}
                   && ($choix eq 'handle_samba_group_modify_struct')
                   && defined($query->param('id'))) {
               my $group = $query->param('id');
	       my $groupes_ref = group_list(\@exclude_groups,
                                            $minimal_gid,
                                            \$error_reason);
               if (list_exists($group, @{$groupes_ref})) {
		  my %param_hash;
		  my @incorrect_fields;
                  generate_samba_group_structure(\%samba_group_modify_struct,
                                                 $group,
                                                 \$error_reason); # err. ign.
		  if (form_validate_input($query,
					  \%samba_group_modify_struct,
					  \%param_hash,
					  \@incorrect_fields)) {
                     my @added_members;
                     my @removed_members;

                     parse_samba_group_structure($group,
                                                 \%param_hash,
                                                 \@added_members,
                                                 \@removed_members);

                     if (group_update_members($group,
                                              \@added_members,
                                              \@removed_members,
                                              \$error_reason)) {
                        $message = 'Groupe mis  jour';
                     }
                     else {
			$result = 1;
			$error_reason = 'Erreur lors du changement: '
                                        . $error_reason;
                     }
		  }
		  else {
		     $result = 1;
		     $error_reason = 'Champs incorrects: '
				     . join(', ', @incorrect_fields);
		  }
               }
               else {
		  $result = 1;
		  $error_reason = 'Groupe inexistant';
               }
            }
            elsif ($session_information{'is_superuser'}
                   && ($choix eq 'groupe_modifier')
                   && defined($query->param('id'))) {
               my $group = $query->param('id');
	       my $groupes_ref = group_list(\@exclude_groups,
                                            $minimal_gid,
                                            \$error_reason);
               if (list_exists($group, @{$groupes_ref})) {
		  $message = "Modification du groupe " . $group;

                  $additional_form_parameters{'id'} = $group;
                  if (generate_samba_group_structure
                         (\%samba_group_modify_struct,
                         $group,
                         \$error_reason)) {
 		     @create_form = ('samba_group_modify_struct');
                  }
                  else {
                     $result = 1;
                  }
               }
               else {
		  $result = 1;
		  $error_reason = 'Groupe inexistant';
               }
            }
	    elsif ($session_information{'is_superuser'}
                   && ($choix eq 'liste_machines')) {
               # all in passwd with special machine group
               # delete
               # then menu to add

               my $title = 'Liste des machines';
	       start_html('text/html', $title, \%context);
	       $printed_header = 1;

               my $machines_ref = machine_list([],
					       $windows_machine_group,
					       $minimal_uid,
					       \$error_reason);
               if (defined($machines_ref)) {
                  my @titles = ('machine');
                  my @content;

                  foreach (sort @{$machines_ref}) {
                     push (@content, { 'machine' => $_ });
                  }

		  my @url_actions
		     = ({ 'descr' => "supprimer",
			  'action_url' => $myself
					  . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=machine_supprimer_c',
			  'id' => 'machine'
			});

		  results_print($title,
				\@content,
				\@titles,
				undef,
				$myself,
				undef,
				undef,
				undef,
				\@url_actions,
				$query);

                  print $query->p,
                        $query->a({href=> $myself
                                          . '?session='
					  . uri_escape($context{'session_id'})
					  . '&commande=machine_ajouter'},
                                  "Ajouter une machine");
               }
               else {
		  $error_reason = 'liste machines impossible: '
				  . $error_reason;
		  $result = 1;
               }
	    }
	 }
      }
      else {
	 if (defined($query->param('mode'))) {
	    my $mode = $query->param('mode');

	    if ($mode eq 'login') {
	       my @incorrect_fields;
	       my %param_hash;
	       if (form_validate_input($query,
				       \%login_struct,
				       \%param_hash,
				       \@incorrect_fields)) {
                  if (smb_verify_password($param_hash{'login'},
                                          $param_hash{'password'})) {
		     $message = 'login correct';

                     $session_information{'identifiant'}
                        = $param_hash{'login'};
                     $session_information{'is_superuser'}
                        = is_superuser($session_information{'identifiant'},
                                       $config_hash{'admin-group'}) ? 't'
                                                                    : 'f';

		     # Drop the unique existing session if any.
		     if (session_drop_selected($dbh,
					       $session_information
                                                  {'identifiant'},
					       \$error_reason)) {

			my %additional_values;
			foreach (@additional_session_information) {
			   $additional_values{$_} = $session_information{$_};
			}

                        if ($session_information{'is_superuser'} eq 't') {
                           $session_information{'is_superuser'} = 1;
                        }
                        else {
                           $session_information{'is_superuser'} = 0;
                        }

			# session_start may also fail (UNIQUE).
			if (!session_start($dbh,
                                           $session_information
                                              {'identifiant'},
					   $session_validity,
					   \$context{'session_id'},
					   \$error_reason,
                                           \%additional_values)) {
			   $error_reason = 'session ne peut tre cre: '
					   . $error_reason;
			   $result = 1;
			}
		     }
		     else {
			$error_reason = 'erreur en effaant de vieilles'
                                        . ' sessions: '
					. $error_reason;
			$result = 1;
		     }

		     if ($result) {
			delete $session_information{'identifiant'};
                        $session_information{'is_superuser'} = 0;
			undef $message;
			delete $context{'session_id'};
		     }
		  }
		  else {
		     $error_reason = 'identification incorrecte';
		     $result = 1;
		  }
	       }
	       else {
		  $result = 1;
		  $error_reason = 'Champs incorrects: '
				  . join(', ', @incorrect_fields);
	       }
	    }
	    else {
	       $error_reason = 'mode incorrect ' . $mode;
	       $result = 1;
	    }
	 }
      }

      if (!$printed_header) {
	 my $title;
	 if (defined($session_information{'identifiant'})) {
	    $title = 'Bienvenue';
            if ($session_information{'is_superuser'}) {
               $title .= ' [administration]';
            }
	 }
	 else {
	    $title = 'Connexion';
	 }

	 start_html('text/html', $title, \%context);
	 $printed_header = 1;
      }

      if (defined($message)) {
	 print $query->p, encode_entities($message);
	 undef $message;
      }

      if ($result) {
	 print $query->h1('Erreur'),
	       $query->p,
	       encode_entities($error_reason);
	 $result = 0;
      }

      my $what;
      if (defined($session_information{'identifiant'})) {
	 if (scalar(@create_form)) {
            $what = '';
            my $create_form;
            while (($create_form = shift @create_form)
                   && exists $the_forms{$create_form})  {
	       my $tmp_what = form_create($query,
					  $the_forms{$create_form},
					  $myself,
					  { 'commande'
                                               => 'handle_' . $create_form,
					    'session'
                                               => $context{'session_id'},
					    %additional_form_parameters
					  },
					  $dont_clean,
					  \$error_reason);
               if (defined($tmp_what) && defined($what)) {
                  $what .= $tmp_what;
               }
               else {
                  undef $what;
               }
            }
	 }
	 else {
	    $what = '';
	 }
      }
      else {
	 $what = menu_create($query,
			     \%login_menu_struct,
			     $myself,
			     {},
			     \$error_reason);
      }

      if (defined($what)) {
	 print $what;
      }
      else {
	 $error_reason = 'erreur  la cration de l\'HTML: '
			 . $error_reason;
	 $result = 1;
      }

      if (defined($session_information{'identifiant'})) {
	 my $active_count = session_active_count($dbh,
						 \$error_reason);
	 if (defined($active_count)) {
	    print $query->p,
		  '[Il y a ',
		  $active_count,
		  ' session',
		  $active_count > 1 ? 's' : '',
                  ' non termine',
		  $active_count > 1 ? 's' : '',
		  ']';
	 }
      }

      if (defined($context{'session_id'})) {
	 if (!session_suspend($dbh,
			      $context{'session_id'},
			      \$error_reason)) {
	    $error_reason = 'session ne peut tre suspendue: '
			    . $error_reason;
	    $result = 1;
	 }
      }

      if (!close_database(\$dbh, \$error_reason)) {
	 $error_reason = 'close_database() error' . $error_reason;
	 $result = 1;
      }
   }
   else {
      $error_reason = 'could not open_database(): ' . $error_reason;
      $result = 1;
   }
}
else {
   $error_reason = 'loading config file '
                   . $config_file
                   . ': '
                   . $error_reason;
   $result = 1;
}

if ($result) {
   if (!$printed_header) {
      start_html('text/html', 'Erreur', undef);
      $printed_header = 1;
   }

   print $query->h1('Erreur'),
         $query->p,
         '<STRONG>Erreur: </STRONG>',
         encode_entities($error_reason);
}

end_html();

sub our_special_menu {
   my ($context_ref, $error_reason_ref) = @_;

   # Errors are mostly ignored.

   my $what = '';
   if ($context_ref->{'session_information'}->{'is_superuser'}) {
      $what .= menu_create($query,
                           $context_ref->{'admin_menu_struct_ref'},
                           $context_ref->{'myself'},
                           { 'session' => $context_ref->{'session_id'} },
                           $error_reason_ref);
   }

   $what .= menu_create($query,
           	        $context_ref->{'user_menu_struct_ref'},
		        $context_ref->{'myself'},
		        { 'session' => $context_ref->{'session_id'} },
		       $error_reason_ref);

   return $what;
}

sub start_html {
   my ($header_type, $title, $context_ref) = @_;

   # HACK HACK HACK
   open(STDOUT, ">&main::OLDSTDOUT"); # err. ign.
   # END HACK HACK HACK

   print $query->header($header_type);
   print $query->start_html(-title => $title);

   cat_file($html_files{'header'});

   if (defined($context_ref) && defined($context_ref->{'session_id'})) {
      my $error_reason;
      my $what = our_special_menu($context_ref, \$error_reason);
      if (defined($what)) {
	 print $what;
      }
   }

   cat_file($html_files{'header-end'});
}

sub end_html {
   cat_file($html_files{'footer'});
   print $query->end_html;
}

sub cat_file {
   my ($file_name) = @_;

   local *FILE;
   if (open(FILE, '<', $file_name)) {
      while (<FILE>) {
         print;
      }

      close(FILE); # err. ign.
   } # else err. ign.
}
   
sub users_list {
   my ($exception_array_ref, $the_gid, $min_uid, $error_reason_ref) = @_;

   my @list;

   setpwent();
   while (my ($name, $passwd, $uid, $gid, $quota, $comment,
              $gcos, $dir, $shell, $expire) = getpwent()) {
      if (($uid >= $min_uid) && ($gid == $the_gid)) {
         push (@list, $name);
      }
   }
   endpwent();

   my @tmp = list_remove($exception_array_ref, @list);
   return \@tmp;
}

sub machine_list {
   my ($exception_array_ref, $the_gid, $min_uid, $error_reason_ref) = @_;

   my $array_ref = users_list($exception_array_ref,
                              $the_gid,
                              $min_uid,
                              $error_reason_ref);

   my @temp;

   foreach my $val (@{$array_ref}) {
      $val =~ s/\$$//;
      push(@temp, $val);
   }

   return \@temp;
}

sub group_list {
   my ($exception_array_ref, $min_gid, $error_reason_ref) = @_;

   my @list;

   setgrent();
   while (my ($name, $passwd, $gid, $members) = getgrent()) {
      if ($gid >= $min_gid) {
         push (@list, $name);
      }
   }
   endgrent();

   my @tmp = list_remove($exception_array_ref, @list);
   return \@tmp;
}

# NOTES
#    - Also returns the system and other groups, this is required so
#      to keep those groups when we change groups.
# BUGS
#    - Not at all optimal
#    - See the BUGS sections of ../NOTES about the way we handle membership.
sub user_get_groups {
   my ($user, $error_reason_ref) = @_;

   my $ok = 1;

   my @list;

   my $groupes_ref = group_list([],
                                0,
                                \$error_reason);
   if (defined($groupes_ref)) {
      foreach my $group (@{$groupes_ref}) {
	 my $group_list_ref = group_list_group([],
					       $group,
					       $error_reason_ref);
	 if (defined($group_list_ref)) {
            if (list_exists($user, @{$group_list_ref}))  {
               push(@list, $group);
            }
         }
         else {
	    $$error_reason_ref = 'cannot list group: ' . $$error_reason_ref;
	    $ok = 0;
         }
      }
   }
   else {
      $$error_reason_ref = 'cannot list groups: ' . $$error_reason_ref;
      $ok = 0;
   }

   if ($ok) {
      return \@list;
   }
   else {
      return undef;
   }
}

sub group_list_group {
   my ($exception_array_ref, $group_name, $error_reason_ref) = @_;

   # BUGS
   #    - error detection weak
   my ($name, $passwd, $gid, $members) = getgrnam($group_name);

   my @list = split(/\s+/, $members);

   my @tmp = list_remove($exception_array_ref, list_unique(@list));
   return \@tmp;
}

sub group_get_gid {
   my ($group_name) = @_;

   # BUGS
   #    - error detection weak
   my ($name, $passwd, $gid, $members) = getgrnam($group_name);

   return $gid;
}

sub delete_user {
   my ($what, $error_reason_ref) = @_;

   my $users_ref = users_list([],
                              $windows_user_group,
                              $minimal_uid,
                              \$error_reason);
   if (list_exists($what, @{$users_ref})) {
      return exec_piped_command($config_hash{'command-manage-smb-user'}
                                . ' del',
				[ $what ],
				$error_reason_ref);
   }
   else {
      $$error_reason_ref = 'invalid user name';
      return 0;
   }
}

sub delete_machine {
   my ($what, $error_reason_ref) = @_;

   my $machine_ref = machine_list([],
				$windows_machine_group,
				$minimal_uid,
				\$error_reason);
   if (list_exists($what, @{$machine_ref})) {
      return exec_piped_command($config_hash{'command-manage-smb-machine'}
                                . ' del',
				[ $what ],
				$error_reason_ref);
   }
   else {
      $$error_reason_ref = 'invalid machine name';
      return 0;
   }
}

sub delete_group {
   my ($what, $error_reason_ref) = @_;

   my $groupes_ref = group_list(\@exclude_groups,
                                $minimal_gid,
                                \$error_reason);
   if (list_exists($what, @{$groupes_ref})) {
      return exec_piped_command($config_hash{'command-manage-smb-group'}
                                . ' del',
				[ $what ],
				$error_reason_ref);
   }
   else {
      $$error_reason_ref = 'invalid group name';
      return 0;
   }
}

# Legality of data is checked upstream, and we pipe it, so
# no risk.
sub user_add {
   my ($user, $password, $error_reason_ref) = @_;

   return exec_piped_command($config_hash{'command-manage-smb-user'}
                             . ' add',
		             [ $user, $password ],
			     $error_reason_ref);
}

sub user_update_password {
   my ($user, $password, $error_reason_ref) = @_;

   return exec_piped_command($config_hash{'command-manage-smb-user'}
                             . ' chg',
		             [ $user,
                               $password,
                               $file_server_admin::no_change_groups ],
			     $error_reason_ref);
   return 0;
}

sub group_add {
   my ($group, $error_reason_ref) = @_;

   return exec_piped_command($config_hash{'command-manage-smb-group'}
                             . ' add',
		             [ $group ],
			     $error_reason_ref);
}

sub machine_add {
   my ($machine, $error_reason_ref) = @_;

   return exec_piped_command($config_hash{'command-manage-smb-machine'}
                             . ' add',
		             [ $machine ],
			     $error_reason_ref);
}

sub generate_samba_group_structure {
   my ($hash_ref, $group, $error_reason_ref) = @_;

   my $group_list_ref = group_list_group([],
                                         $group,
                                         $error_reason_ref);
   if (defined($group_list_ref)) {
      my $count = 1;
      foreach (sort @{$group_list_ref}) {
         $hash_ref->{$count}
            = { 'name'  => 'user_member_' . $_,
                'descr' => $_,
                'type'  => 'boolean',
                'check' => \&form_valid_param,
                'optional' => undef,
                'preset' => 1
              };
         $count++
      }
   
      my $users_ref = users_list([],
                                 $windows_user_group,
                                 $minimal_uid,
                                 $error_reason_ref);
      if (defined($users_ref)) {
         foreach (sort @{$users_ref}) {
            if (!list_exists($_, @{$group_list_ref})) {
               $hash_ref->{$count}
                  = { 'name'  => 'user_member_' . $_,
                      'descr' => $_,
                      'type'  => 'boolean',
                      'check' => \&form_valid_param,
                      'optional' => undef,
                      'preset' => 0
                    };
               $count++
            }
        }
      }                     
   }
   else {
      $$error_reason_ref = 'liste du groupe impossible: '
                      . $$error_reason_ref;
      return 0;
   }

   return 1;
}

sub parse_samba_group_structure {
   my ($group,
       $param_hash_ref,
       $added_members_array_ref,
       $removed_members_array_ref) = @_;

   my $error_reason;

   my $group_list_ref = group_list_group([],
                                         $group,
                                         \$error_reason);
   if (defined($group_list_ref)) {
      foreach my $param (keys %{$param_hash_ref}) {
	 if ($param =~ /^user_member_(.+)$/) {
            my $user = $1;
	    if (list_exists($user, @{$group_list_ref})) {
               if (!$param_hash_ref->{$param}) {
                  push(@{$removed_members_array_ref}, $user);
               }
            }
            else {
               if ($param_hash_ref->{$param}) {
                  push(@{$added_members_array_ref}, $user);
               }
            }
         }
      }
   }
   # else err. ign.
}

# BUGS
#    - Not at all optimal
sub group_update_members {
   my ($group,
       $added_members_array_ref,
       $removed_members_array_ref,
       $error_reason_ref) = @_;

   my $ok = 1;

   foreach my $user (@{$added_members_array_ref}) {
      my $groups_ref = user_get_groups($user, $error_reason_ref);
      if (defined($groups_ref)) {
	 push (@{$groups_ref}, $group); # add the group

	 if (!user_update_groups($user, $groups_ref, $error_reason_ref)) {
	    $ok = 0;
	    $$error_reason_ref = 'issue while setting groups for user '
                                 . $user;
         }
      }
      else {
         $ok = 0;
         $$error_reason_ref = 'issue while getting groups for user ' . $user;
      }
   }

   foreach my $user (@{$removed_members_array_ref}) {
      my $groups_ref = user_get_groups($user, $error_reason_ref);
      if (defined($groups_ref)) {
         my @new_groups = list_remove([ $group ], @{$groups_ref});

	 if (!user_update_groups($user, \@new_groups, $error_reason_ref)) {
	    $ok = 0;
	    $$error_reason_ref = 'issue while setting groups for user '
                                 . $user;
         }
      }
      else {
         $ok = 0;
         $$error_reason_ref = 'issue while getting groups for user ' . $user;
      }
   }

   return $ok;
}

sub user_update_groups {
   my ($user, $groups_ref, $error_reason_ref) = @_;

   return exec_piped_command($config_hash{'command-manage-smb-user'}
                             . ' chg',
		             [ $user,
                               '',
                               join(' ', @{$groups_ref}),
                             ],
			     $error_reason_ref);
}

# BUGS
#    - see ../scripts/manage-smb
sub form_valid_user_machine_or_group {
   my ($hash_ref, $param) = @_;

   return $param =~ /^([a-zA-Z][a-zA-Z0-9\-_]*)$/;
}

# NOTES
#    - an error means not superuser, not told
sub is_superuser {
   my ($login, $admin_group) = @_;

   my $error_reason; # ignored

   my $is_superuser = 0;

   my $users_ref = group_list_group([],
                                    $admin_group,
                                    \$error_reason);
   if (defined($users_ref)) {
      if (list_exists($login, @{$users_ref})) {
         $is_superuser = 1;
      }
   }

   return $is_superuser;
}
